// $Id: updatepos.c,v 1.2 2000/12/29 11:08:48 hip Exp $

#include <math.h>
#include "xgrav.h"

// reference globals declared in xgrav.c
extern Planet a, b, c;
extern unsigned int dT;
extern double crashdist2;
extern double newton_gamma;


int updatePlanet(Planet *a, Planet *b, Planet *c)
{
	static double Ts, r2, F1, F1x, F1y, F2, F2x, F2y, Fx, Fy;
	static double dx, dy, alpha, ax, ay; 

	Ts = dT;

	if(!a->fixed){
		// force component toward planet b
		dx = b->p.x - a->p.x;								// co-axial distance components
		dy = b->p.y - a->p.y;
		alpha = atan2(dy, dx);							// aspect angle
		r2 = dx * dx + dy * dy;							// distance between planets
		if(r2 < crashdist2)
			return 0;
		F1 = newton_gamma * (a->m * b->m) / r2; // Newton's gravitational law
		F1x = F1 * cos(alpha);							// co-axial force components
		F1y = F1 * sin(alpha);

		// force component toward planet c
		dx = c->p.x - a->p.x;								// co-axial distance components
		dy = c->p.y - a->p.y;
		alpha = atan2(dy, dx);							// aspect angle
		r2 = dx * dx + dy * dy;							// distance between planets
		if(r2 < crashdist2)
			return 0;
		F2 = newton_gamma * (a->m * c->m) / r2; // Newton's gravitational law
		F2x = F2 * cos(alpha);							// co-axial force components
		F2y = F2 * sin(alpha);

		Fx = F1x + F2x;											// total force vector components
		Fy = F1y + F2y;

		ax = Fx / a->m;											// acceleration components
		ay = Fy / a->m;

		a->v.x = a->v.x + ax * Ts;					// vt = v0 + a*t
		a->v.y = a->v.y + ay * Ts;
	
		// st = s0 + v0 * t + 1/2 * a * t^2
		a->p.x = a->p.x + a->v.x * Ts + 0.5 * ax * Ts * Ts; 
		a->p.y = a->p.y + a->v.y * Ts + 0.5 * ay * Ts * Ts;
	}
	return 1;
}


int updatepos()
{
	Planet aa, bb, cc;
	int p, q, r;

	aa = a; bb = b; cc = c;
	p = updatePlanet(&aa, &b, &c);
	q = updatePlanet(&bb, &a, &c);
	r = updatePlanet(&cc, &a, &b);
	a = aa; b = bb; c = cc;

	return p && q && r;
}
