// $Id: xgrav.c,v 1.2 2000/12/29 11:08:48 hip Exp $
//
// GRAVITY - A simulation based on Newton's General Law of Gravitation
// http://www.xs4all.nl/~hipster/lib/c++
//
// This code was one of my first attempts at programming X, some very long
// time ago. This is *not* how you'd want to do it  ;)
//
// TODO
// - screen backup bitmap to prevent image-loss, or store calculated coords
//   in a file >> store coords in memory rather than in a file.
//   use a memory mapped file for the log, or even just an array
// - generalise calculations to arbitrary (>1) number of masses
// - add 3rd dimension to calculation model and viewing system
// - improve collission-detection. Masses sometimes veer off after a collission
// - setup-mode for positioning masses and setting their properties
// - make all internal parameters and thresholds accessible via UI 
// - get rid of xgrav.rc-file
// - base planet color on speed (grayscale)
// - query the display for color depth and adjust coloring appropiately
// - fix mouse move/resize stuff, it seriously damaged now
//
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <math.h>
#include <time.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/keysym.h>
#include "xgrav.h"

char *progname = "Gravity";
char *displayname = NULL;

Display *dpy;
int scr;
Window win;
GC gc;
XSizeHints sh;
XWindowAttributes wa;
XEvent event;
KeySym ksym;
unsigned long bdcolor, bgcolor;
int bwidth = 2;

Planet a, b, c;
double scale;														// display scale factor
double newton_gamma = 6.67e-11;					// gravitational constant
int psize = 4;
double crashdist2 = 5e6 * 5e6;					// (sqr) minimal distance betw. planets
unsigned int T = 0, dT, drep;						// time, timestep, reporting interval
int fStats = 1;
int winx, winy;													// window size
int cx = 0, cy = 0;											// coordinate centre
time_t tstart;
int collision = 0;

void readparams(char*);
void stats();
void drawproc();
void drawplanets();
void drawplanet(Vector, int);
extern int updatepos();
void trans2screen(int*, int*);
int handleEvent(XEvent*);
void freeXstuff();
void drawlog();
FILE *fplog;


int main(int argc, char **argv)
{
	//XGCValues xgcv;
	
	readparams(argc != 2 ? "xgrav.rc" : argv[1]);

	if((dpy = XOpenDisplay(displayname)) == NULL){
		fprintf(stderr, "%s: Failed to open display.\n", progname);
		exit(1);
	}
	
	scr = DefaultScreen(dpy);

	// make a window
	sh.x = sh.y = 0;
	sh.width = 512;
	sh.height = 512;
	sh.flags = PPosition | PSize;

	win = XCreateSimpleWindow(
		dpy,
		RootWindow(dpy, scr),
		sh.x, sh.y,
		sh.width, sh.height,
		bwidth, bdcolor, bgcolor);

	bdcolor = BlackPixel(dpy, scr);
	bgcolor = WhitePixel(dpy, scr);
	//bgcolor = BlackPixel(dpy, scr); 
	XSetWindowBackground(dpy, win, bgcolor);
	XClearWindow(dpy, win);
						
	XSetStandardProperties(dpy, win, progname, progname, None, NULL, 0, &sh);
	XSelectInput(dpy, win,
		ExposureMask |
		ButtonPressMask |
		ButtonReleaseMask |
		ButtonMotionMask |
		KeyPressMask);
		//StructureNotifyMask|VisibilityChangeMask
	XMapWindow(dpy, win);
	XFlush(dpy);

	gc = XCreateGC(dpy, win, 0, 0);
	
	//xgcv.function = GXset;
	//XChangeGC(dpy, gc, GCFunction, &xgcv);

	// RGB-555 color model
	//a.color = 0x1f;
	//b.color = 0x10 << 5;
	//c.color = 0x1f << 10;
	// RGB-888 color model
	a.color = 0xff;
	b.color = 0x80 << 8;
	c.color = 0xff << 16;

	fplog = fopen("log", "w+");
	tstart = time(NULL);
	drawproc();
	stats();
	if(collision) // allow for some viewing pleasure
		while(42){
			XNextEvent(dpy, &event); // blocks
			if(!handleEvent(&event))
				break;
		}
	fclose(fplog);
	freeXstuff();
	return 0;
}


int handleEvent(XEvent *event)
{
	switch(event->type){
		case ButtonRelease:
			switch(event->xbutton.button){
				// TODO image center dragging
				case Button1:{
					// FIXME incorporate scaling into calculation
					int x, y;
					//printf("old c: %d %d\n", cx, cy);
					x = event->xbutton.x;
					y = event->xbutton.y;
					//printf("click: %d %d\n", x, y);
					cx += x - (winx / 2);
					cy += (winy / 2) - y;
					//printf("new c: %d %d\n", cx, cy);
					drawlog();
					break;
				}
			}
			break;

		//case MotionNotify:
		//	printf("motion %d %d\n", event->xmotion.x, event->xmotion.y);
		//	break;

		case KeyPress:
			ksym = XLookupKeysym((XKeyEvent*)event, 0);
			switch(ksym){ // see X11/keysymdef.h for keys
				case XK_q: return 0;
				case XK_a: scale /= 2; drawlog(); break;
				case XK_z: scale *= 2; drawlog(); break;
			}
			break;

		case Expose:
			// FIXME a window maximize still causes multiple redraws, restoring it
			// to normal size after that doesn't, however...
			if(event->xexpose.count == 0){
				//XClearWindow(dpy, win);
				XGetWindowAttributes(dpy, win, &wa);
				winx = wa.width;
				winy = wa.height;
				drawlog();
			}
			break;
	}
	return 1;
}


void freeXstuff()
{
	XFreeGC(dpy, gc);
	XDestroyWindow(dpy, win);
	XCloseDisplay(dpy);
}


void drawlog()
{
	Vector v[3];

	XClearWindow(dpy, win);
	fseek(fplog, 0, SEEK_SET);
	while(!feof(fplog)){
		fread(v, sizeof(Vector), 3, fplog);
		drawplanet(v[0], a.color);
		drawplanet(v[1], b.color);
		drawplanet(v[2], c.color);
	}
	fseek(fplog, 0, SEEK_END);
}


void readparams(char *rc)
{
	FILE *fp;

	if(fp = fopen(rc, "rt"), !fp){
		printf("Cannot open '%s'\n", rc);
		exit(1);
	}

	fscanf(fp, "%lf", &scale);
	fscanf(fp, "%d %d", &dT, &drep);
	fscanf(fp, "%lf %lf", &a.p.x, &a.p.y);
	fscanf(fp, "%lf %lf", &a.v.x, &a.v.y);
	fscanf(fp, "%lf", &a.m);
	fscanf(fp, "%d", &a.fixed);
	fscanf(fp, "%lf %lf", &b.p.x, &b.p.y);
	fscanf(fp, "%lf %lf", &b.v.x, &b.v.y);
	fscanf(fp, "%lf", &b.m);
	fscanf(fp, "%d", &b.fixed);
	fscanf(fp, "%lf %lf", &c.p.x, &c.p.y);
	fscanf(fp, "%lf %lf", &c.v.x, &c.v.y);
	fscanf(fp, "%lf", &c.m);
	fscanf(fp, "%d", &c.fixed);

	fclose(fp);
}


void s2ydhms(long t)
{
	int y, d, h, m;

	y = t / (60 * 60 * 24 * 365);
	t = t - (60 * 60 * 24 * 365) * y;
	d = t / (60 * 60 * 24);
	t = t - (60 * 60 * 24) * d;
	h = t / (60 * 60);
	t = t - (60 * 60) * h;
	m = t /  60;
	t = t -  60 * m;

	printf("%03dy %03dd %02dh %02dm %02lds\n", y, d, h, m, t);
}


void stats()
{
	unsigned int tend, tdiff;
	tend = time(NULL);
	tdiff = tend - tstart;
	
	printf("Simulated time  : %12ds | ", T);
	s2ydhms(T);
	
	printf("Real time       : %12ds | ", tdiff);
	s2ydhms(tdiff);

	printf("Time compression: %g s/s\n", (double)T / tdiff);
}


void drawproc()
{
	XEvent ev;
  int rep;
	char msg[0x100];

  while(42){
    T += dT;
    rep = fStats && !(T % drep);

    if(updatepos()){
      if(rep){
				fwrite(&a.p, sizeof(Vector), 1, fplog);
				fwrite(&b.p, sizeof(Vector), 1, fplog);
				fwrite(&c.p, sizeof(Vector), 1, fplog);
        drawplanets();
				sprintf(msg, "a.v:%4.0f b.v:%4.0f c.v:%4.0f",
					sqrt(a.v.x*a.v.x + a.v.y*a.v.y),
					sqrt(b.v.x*b.v.x + b.v.y*b.v.y),
					sqrt(c.v.x*c.v.x + c.v.y*c.v.y)
				);
				XClearArea(dpy, win, 0, 0, 200, 20, False);
				XDrawString(dpy, win, gc, 4, 16, msg, strlen(msg));
				XFlush(dpy);
			}
		}
    else{
			sprintf(msg, "*** collision ***");
      puts(msg);
			XDrawString(dpy, win, gc, 4, 30, msg, strlen(msg));
			collision = 1;
      return;
    }

		if(XPending(dpy)){ //non-blocking check for XEvents
			XNextEvent(dpy, &ev);
	  	if(!handleEvent(&ev))
				return;
		}
  }
}


void drawplanets()
{
	drawplanet(a.p, a.color);
	drawplanet(b.p, b.color);
	drawplanet(c.p, c.color);
}


void drawplanet(Vector p, int c)
{
  int x, y;
  
  x = p.x / scale;
  y = p.y / scale;
  trans2screen(&x, &y);
	XSetForeground(dpy, gc, c);
	XDrawPoint(dpy, win, gc, x, y);
}


void trans2screen(int *x, int *y)
{
  *x += (winx / 2) - cx;
  *y = (winy / 2) - *y + cy;
}
